/*

  gb2pdb  -  Convert a gameboy game image into a palm database
             for use with PalmBoy

	   Version 1.0 - Aug 1999 Till Harbaum (T.Harbaum@tu-bs.de)
	   Version 1.1 - June 2000 Bodo Wenzel (bodowenzel@web.de)
	   Version 1.2 - September 2000 Bodo Wenzel (bodowenzel@web.de)
*/

/* Layout of database:

record-#   size         content
-----------------------------------------------------
0          32768        ROM banks 0 and 1 (same for banked and unbanked)
1..2^n     16384        ROM banks 2 to last (if applicable)
2^n+1      8192..32768  RAM banks 0 to 3 (if applicable)

A cartridge with RAM can be determined by the number of records.

Examples:
  Simple 32K cartridge:
    number of records = 1, only record 0
  Cartridge with 64K ROM and 32 K RAM: 
    number of records = 4, record 0 (32K ROM), 1 (16K ROM), 2 (16K ROM), 3 (32K RAM)

*/

#define DB_CREATOR  "PBoy"
#define DB_TYPE     "GAME"

#define ROM_BANK      16384U
#define RAM_BANK      8192U
#define RAM_MAX_SIZE  32768U

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned char ubyte;
typedef signed char sbyte;
typedef unsigned short uword;
typedef short word;
typedef unsigned long udword;
typedef long dword;

typedef struct {
  ubyte name[32];
  uword fileAttributes;
  uword version;
  udword creationDate;
  udword modificationDate;
  udword lastBackupDate;
  udword modificationNumber;
  udword appInfoArea;
  udword sortInfoArea;
  ubyte databaseType[4];        
  ubyte creatorID[4];
  udword uniqueIDSeed;
  udword nextRecordListID;
  uword numberOfRecords;
}
#ifdef __GNUC__
  __attribute__ ((packed))
#endif
  PdbHeader;

typedef struct {
  udword recordDataOffset;
  ubyte recordAttributes;
  ubyte uniqueID[3];
}
#ifdef __GNUC__
  __attribute__ ((packed))
#endif
  RecHeader;

void 
WriteWord(void *addr, uword data) {
  ubyte *to = (ubyte*)addr;

  to[0] = (data>>8)&0xff; 
  to[1] = (data>>0)&0xff;
}

void 
WriteDWord(void *addr, udword data) {
  ubyte *to = (ubyte*)addr;

  to[0] = (data>>24)&0xff; 
  to[1] = (data>>16)&0xff;
  to[2] = (data>>8)&0xff;
  to[3] = (data>>0)&0xff;
}

char *mbc_str[]={
  "ROM ONLY",
  "ROM+MBC1",
  "ROM+MBC1+RAM",
  "ROM+MBC1+RAM+BATTERY",
  "",
  "ROM+MBC2",
  "ROM+MBC2+BATTERY"
};

char gb_buffer[RAM_MAX_SIZE];

int
main(int argc, char **argv) {
  char *p,*outname;
  int i;
  FILE *outfile, *infile;
  size_t j;
  PdbHeader pdb_header;
  RecHeader rec_header;
  unsigned long offset;
  unsigned char gb_info[0x50];
  int ram_banks[]={0,1,1,4}, ram_size[]={0,2,8,32};
  int rom_banks, nrecs;
  unsigned short filler;

  printf("gb2pdb V1.0      (c) 1999 by Till Harbaum\n"
         "       V1.1-V1.2 (c) 2000 by Bodo Wenzel\n");

  if((argc!=2)&&(argc!=3)&&(argc!=4)) {
    puts("usage: gb2pdb infile [outfile [dbname]]");
    return 1;
  }

  infile=fopen(argv[1], "rb");
  if(infile==NULL) {
    printf("error opening input file '%s'\n", argv[1]);
    return 1;
  }

  /* create destination file name */
  if(argc==2) {
    outname=(char*)malloc(strlen(argv[1])+4+1);
    if (outname==NULL) {
      printf("failed allocating memory\n");
      return 1;
    }
    strcpy(outname,argv[1]);
    if((p=strrchr(outname,'.'))!=NULL) {
      strcpy(p,".pdb");
    } else
      strcat(outname,".pdb");
  } else {
    outname=argv[2];
  }

  /* open destination file */
  outfile=fopen(outname,"wb");
  printf("writing %s\n", outname);
  if(outfile==NULL) {
    printf("error opening output file\n");
    return 1;
  }

  /* scan some image values */
  fseek(infile, 0x100l, SEEK_SET);
  fread(gb_info, 0x50, 1, infile);
  fseek(infile, 0l, SEEK_SET);

  rom_banks = 2 << gb_info[0x48];

  if((gb_info[0x47]==4)||(gb_info[0x47]>6)) {
    printf("unknown/unsupported MBC %d\n", gb_info[0x47]);
    fclose(infile);
    fclose(outfile);
    return 1;
  }

  /* clean up cartridge name */
  if (argc!=4) {
    if (gb_info[0x43]==0xc0) {
      printf("warning: color-only game\n");
      for (i=0x34; i<0x44-5; i++) {
        if (gb_info[i]=='\0')
          break;
        if (!isprint(gb_info[i]))
          gb_info[i]=' ';
      }
    } else if (gb_info[0x43]==0x80) {
      /* game with color and gray support */
      for (i=0x34; i<0x44-1; i++) {
        if (gb_info[i]=='\0')
          break;
        if (!isprint(gb_info[i]))
          gb_info[i]=' ';
      }
    } else {
      /* old game, gray-only */
      for (i=0x34; i<0x44; i++) {
        if (gb_info[i]=='\0')
          break;
        if (!isprint(gb_info[i]))
          gb_info[i]=' ';
      }
    }
  } else {
    if (gb_info[0x43]==0xc0)
      printf("warning: color-only game\n");
    for (i=0x34; i<0x44; i++) {
      gb_info[i]=argv[3][i-0x34];
      if (gb_info[i]=='\0')
        break;
    }
  }
  gb_info[i]='\0';

  printf("Name:     %s (%s)\n", &gb_info[0x34], mbc_str[gb_info[0x47]]);
  printf("ROM size: %dkB (%d banks)\n", (int)(rom_banks*(ROM_BANK/1024)), rom_banks);
  printf("RAM size: %dkB (%d banks)\n", ram_size[gb_info[0x49]], ram_banks[gb_info[0x49]]); 

  nrecs=rom_banks - 1;
  if (ram_banks[gb_info[0x49]]!=0)
    nrecs++;

  /* create/write pdb header */
  memset(&pdb_header,0,sizeof(PdbHeader));
  strcpy((char*)pdb_header.name, (char*)(&gb_info[0x34]));
  WriteWord(&(pdb_header.version),0x0001);
  WriteDWord(&(pdb_header.creationDate), 0xb3d1b04dL);
  WriteDWord(&(pdb_header.modificationDate), 0xb3d1b04dL);
  strncpy(pdb_header.databaseType, DB_TYPE, 4);
  strncpy(pdb_header.creatorID, DB_CREATOR, 4);
  WriteWord(&(pdb_header.numberOfRecords), nrecs);
  fwrite(&pdb_header,sizeof(PdbHeader),1,outfile);

  /* offset to data inside db */
  offset = sizeof(PdbHeader) + nrecs * sizeof(RecHeader) + 2;

  /* write record headers */

  WriteDWord(&(rec_header.recordDataOffset), offset);
  rec_header.recordAttributes = 0x60;
  rec_header.uniqueID[0] = rec_header.uniqueID[1] = rec_header.uniqueID[2] = 0x00;
  fwrite(&rec_header,sizeof(RecHeader),1,outfile);
  offset += 2*ROM_BANK;

  for(i=2;i<rom_banks;i++) {
    WriteDWord(&(rec_header.recordDataOffset), offset);
    rec_header.recordAttributes = 0x60;
    rec_header.uniqueID[0] = rec_header.uniqueID[1] = rec_header.uniqueID[2] = 0x00;
    fwrite(&rec_header,sizeof(RecHeader),1,outfile);
    offset += ROM_BANK;
  }

  if(ram_banks[gb_info[0x49]]!=0) {
    WriteDWord(&(rec_header.recordDataOffset), offset);
    rec_header.recordAttributes = 0x60;
    rec_header.uniqueID[0] = rec_header.uniqueID[1] = rec_header.uniqueID[2] = 0x00;
    fwrite(&rec_header,sizeof(RecHeader),1,outfile);
  }

  filler = 0;
  fwrite(&filler,sizeof(filler),1,outfile);

  /* now copy rom content */
  for(i=0;i<rom_banks;i++) {
    if((j=fread(gb_buffer, 1, ROM_BANK, infile))!=ROM_BANK) {
      printf("read error in game image (%d,%ld)\n", i,(long)j);
      fclose(outfile);
      fclose(infile);
      return 1;
    }

    fwrite(gb_buffer, ROM_BANK, 1, outfile);
  }

  /* now create empty ram */
  if (ram_banks[gb_info[0x49]]!=0) {
    j=fread(gb_buffer, 1, RAM_MAX_SIZE, infile);
    /* errors silently ignored */
    memset(gb_buffer+j, 0x00, RAM_MAX_SIZE-j);
    for(i=0;i<ram_banks[gb_info[0x49]];i++)
      fwrite(gb_buffer+i*RAM_BANK, RAM_BANK, 1, outfile);
  }

  if (ferror(outfile)) {
    printf("write error in palm database\n");
    fclose(outfile);
    fclose(infile);
    return 1;
  }

  fclose(outfile);
  fclose(infile);

  return 0;
}
